/**
 * Copyright 1998-2007, CHISEL Group, University of Victoria, Victoria, BC, Canada.
 * All rights reserved.
 */
package ca.uvic.csr.shrimp.usercontrols;

import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import ca.uvic.csr.shrimp.AttributeConstants;
import ca.uvic.csr.shrimp.LayoutConstants;
import ca.uvic.csr.shrimp.DisplayBean.DisplayBean;
import ca.uvic.csr.shrimp.DisplayBean.DisplayConstants;
import ca.uvic.csr.shrimp.DisplayBean.layout.Layout;
import ca.uvic.csr.shrimp.DisplayBean.layout.TreeMapLayout;
import ca.uvic.csr.shrimp.ShrimpApplication.ApplicationAccessor;
import ca.uvic.csr.shrimp.ShrimpApplication.ShrimpApplication;
import ca.uvic.csr.shrimp.ShrimpProject.ShrimpProject;
import ca.uvic.csr.shrimp.ShrimpTool.ShrimpTool;
import ca.uvic.csr.shrimp.gui.quickview.ExpressViewConfigurator;
import ca.uvic.csr.shrimp.gui.quickview.QuickViewAction;

/**
 * Sets up the {@link ExpressViewConfigurator} to show a nested view using the treemap layout.
 * 
 * @author Rob Lintern, Chris Callendar
 */
public class NestedTreemapViewAction extends QuickViewAction {
	
    public static final String ACTION_NAME = "Nested Treemap";

    /**
	 * Don't use this constructor.  It is used to instantiate this class using reflection.
	 */
	public NestedTreemapViewAction() {
		super();
	}

    public NestedTreemapViewAction(ShrimpProject project) {
        super(ACTION_NAME, "icon_quick_view_treemap.gif", project);
        
        config.setArcTypesOfInterest(Collections.EMPTY_LIST);
        config.setNodeTypesOfInterest(getAllNodeTypes());
        config.setLabelMode(DisplayConstants.LABEL_MODE_SCALE_BY_NODE_SIZE);
        config.setLayoutMode(LayoutConstants.LAYOUT_TREEMAP);
        config.setCprels(getDefaultCprels());
        config.setInverted(getDefaultCprelsInverted());
    }
    
    public void startAction() {
        ShrimpApplication app = ApplicationAccessor.getApplication();
        if (!hasProject())
            return;

        DisplayBean displayBean = null;
        ShrimpTool shrimpView = null;
		try {
			shrimpView = getTool(ShrimpProject.SHRIMP_VIEW);
            displayBean = (DisplayBean) shrimpView.getBean(ShrimpTool.DISPLAY_BEAN);
        } catch (Exception e) {
            e.printStackTrace();
            return;
        }
        
        List layouts = displayBean.getLayouts();
        TreeMapLayout treeMapLayout = null;
        for (Iterator iter = layouts.iterator(); iter.hasNext();) {
            Layout layout = (Layout) iter.next();
            if (layout instanceof TreeMapLayout) {
                treeMapLayout = (TreeMapLayout) layout;
                break;
            }
        }
        if (treeMapLayout == null)
            return;
        
        treeMapLayout.setMapLayoutName(TreeMapLayout.LAYOUT_ORDERED);
        treeMapLayout.setOrderField(AttributeConstants.ORD_ATTR_NUM_DESCENDENTS);
        treeMapLayout.setSizeField(AttributeConstants.ORD_ATTR_NUM_DESCENDENTS);
        treeMapLayout.setBorderField(0);
        
        String oldDefaultLayoutMode = app.getProperties().getProperty(DisplayBean.PROPERTY_KEY__DEFAULT_LAYOUT_MODE, DisplayBean.PROPERTY_DEFAULT_VALUE__DEFAULT_LAYOUT_MODE);
        app.getProperties().setProperty(DisplayBean.PROPERTY_KEY__DEFAULT_LAYOUT_MODE, LayoutConstants.LAYOUT_TREEMAP);
        
        // configure the express view
        super.startAction();
		
        OpenAllAdapter openAllAdapter = new OpenAllAdapter(getProject(), shrimpView);
		openAllAdapter.startAction();
		app.getProperties().setProperty(DisplayBean.PROPERTY_KEY__DEFAULT_LAYOUT_MODE, oldDefaultLayoutMode);
    }
 
}
